
import logging
from functools import lru_cache

from y.contracts import Contract, has_methods
from y.decorators import log
from y.prices import magic
from y.utils.multicall import fetch_multicall

logger = logging.getLogger(__name__)

@log(logger)
@lru_cache
def is_mstable_feeder_pool(address: str) -> bool:
    return has_methods(address, ['getPrice()((uint,uint))','mAsset()(address)'])

@log(logger)
def get_price(address, block=None):
    contract = Contract(address)
    ratio, masset, decimals = fetch_multicall([contract,'getPrice'],[contract,'mAsset'],[contract,'decimals'],block=block)
    ratio = ratio[0] / 10 ** decimals
    underlying_price = magic.get_price(masset,block)
    return underlying_price * ratio



import math
from itertools import cycle

from brownie import chain
from eth_abi.packed import encode_abi_packed
from y.classes.common import ERC20
from y.classes.singleton import Singleton
from y.constants import usdc, weth
from y.contracts import Contract, contract_creation_block
from y.exceptions import UnsupportedNetwork
from y.networks import Network
from y.utils.multicall import fetch_multicall

# https://github.com/Uniswap/uniswap-v3-periphery/blob/main/deploys.md
UNISWAP_V3_FACTORY = '0x1F98431c8aD98523631AE4a59f267346ea31F984'
UNISWAP_V3_QUOTER = '0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'

# same addresses on all networks
addresses = {
    Network.Mainnet: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Arbitrum: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000],
    },
}

FEE_DENOMINATOR = 1_000_000


class UniswapV3(metaclass=Singleton):
    def __init__(self):
        if chain.id not in addresses:
            raise UnsupportedNetwork('compound is not supported on this network')

        conf = addresses[chain.id]
        self.factory = Contract(conf['factory'])
        self.quoter = Contract(conf['quoter'])
        self.fee_tiers = conf['fee_tiers']

    def __contains__(self, asset):
        return chain.id in addresses

    def encode_path(self, path):
        types = [type for _, type in zip(path, cycle(['address', 'uint24']))]
        return encode_abi_packed(types, path)

    def undo_fees(self, path):
        fees = [1 - fee / FEE_DENOMINATOR for fee in path if isinstance(fee, int)]
        return math.prod(fees)

    def get_price(self, token: str, block=None):
        if block and block < contract_creation_block(UNISWAP_V3_QUOTER):
            return None

        paths = []
        if token != weth:
            paths += [
                [token, fee, weth.address, self.fee_tiers[0], usdc.address] for fee in self.fee_tiers
            ]

        paths += [[token, fee, usdc.address] for fee in self.fee_tiers]

        results = fetch_multicall(
            *[
                [self.quoter.address, 'quoteExactInput', self.encode_path(path), ERC20(token).scale]
                for path in paths
            ],
            block=block,
        )

        outputs = [
            amount / self.undo_fees(path) / 1e6
            for amount, path in zip(results, paths)
            if amount
        ]
        return max(outputs) if outputs else None


uniswap_v3 = None
try:
    uniswap_v3 = UniswapV3()
except UnsupportedNetwork:
    pass



import logging
from functools import lru_cache
from typing import Union

import brownie
from eth_typing.evm import Address
from y.balancer.balancer import balancer
from y.chainlink.chainlink import chainlink
from y.constants import STABLECOINS
from y.contracts import Contract
from y.curve.curve import curve
from y.decorators import log
from y.prices import (belt, cream, ellipsis, froyo, gelato, ib, mooniswap,
                      mstablefeederpool, piedao, saddle, tokensets, wsteth,
                      yearn)
from y.prices.aave import aave
from y.prices.compound import compound
from y.prices.synthetix import synthetix
from y.uniswap.uniswap import uniswap

logger = logging.getLogger(__name__)

@log(logger)
@lru_cache(maxsize=None)
def check_bucket(
    token_address: Union[str, Address, brownie.Contract, Contract]
    ):

    if type(token_address) != str:
        token_address = str(token_address)

    # these require neither calls to the chain nor contract initialization
    if token_address == "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE":       return 'wrapped gas coin'
    elif token_address in STABLECOINS:                                      return 'stable usd'

    elif wsteth.is_wsteth(token_address):                                   return 'wsteth'
    elif cream.is_creth(token_address):                                     return 'creth'
    elif belt.is_belt_lp(token_address):                                    return 'belt lp'

    elif froyo.is_froyo(token_address):                                     return 'froyo'
    elif aave.is_atoken(token_address):                                     return 'atoken' 

    # these just require calls
    elif balancer.is_balancer_pool(token_address):                          return 'balancer pool'
    elif yearn.is_yearn_vault(token_address):                               return 'yearn or yearn-like'
    elif ib.is_ib_token(token_address):                                     return 'ib token'

    elif gelato.is_gelato_pool(token_address):                              return 'gelato'
    elif piedao.is_pie(token_address):                                      return 'piedao lp'
    elif tokensets.is_token_set(token_address):                             return 'token set'

    elif ellipsis.is_eps_rewards_pool(token_address):                       return 'ellipsis lp'
    elif mstablefeederpool.is_mstable_feeder_pool(token_address):           return 'mstable feeder pool'
    elif saddle.is_saddle_lp(token_address):                                return 'saddle'

    # these require both calls and contract initializations
    elif uniswap.is_uniswap_pool(token_address):                            return 'uni or uni-like lp'
    elif mooniswap.is_mooniswap_pool(token_address):                        return 'mooniswap lp'
    elif token_address in compound:                                         return 'compound'
    elif token_address in curve:                                            return 'curve lp'
    elif token_address in chainlink:                                        return 'chainlink feed'
    elif token_address in synthetix:                                        return 'synthetix'
